Unit BPRename;

{ Class TBPRename }
{ Modified  06-08-02 }

Interface

Uses Windows,SysUtils,BPDef,Debug;

Type TBPRename=class
      private
       FTree:PBPTree;
       FNodesCount:integer;
       FLinksCount:integer;
       FCurLevel:LongInt;
       FCurSSLevel:ShortString;
       FParentLevel:LongInt;
       FNodes:PBPNodeRenameStructure;
       FLinks:PBPLinkRenameStructure;
       FMinLevel:LongInt;
       FMaxLevel:LongInt;
       procedure InitializeNodes(nodes:PBPNode;ncount:integer);
       procedure FinalizeNodes;
       procedure InitializeLinks(links:PBPLink;lcount:integer);
       procedure FinalizeLinks;
       function    GetMultipleMask(index:Integer):TBPMultipleMask;
       function    IsMultipleFirst(index:Integer;MultipleFrom:Boolean):Boolean;
       function    GetTreeLevel(level:ShortString):LongInt;
       function    GetMaxLevel:LongInt;
       function    GetMinLevel(MaxLevel:LongInt):LongInt;
       function    GetNode(NodeNumber:DWORD):PBPNodeRenameStructure;
       function    GetLink(index:integer):PBPLinkRenameStructure;
       function    GetLocalNumber(Name:DWORD):TVeryShortString;
       function    GetLinkTypeFrom(Flags:DWORD):TLinkType;
       function    GetLinkTypeTo(Flags:DWORD):TLinkType;
       function    GetNextLevel(NodeNumber:DWORD):String;
       function    GetCheckLT(lt:TLinkType):TLinkType;
       procedure   RenameGlobalLinks(level:String;oldName,newName:DWORD;FromDir:Boolean;check_lt:TLinkType);
       procedure   RenameMultipleLinks(index:DWORD;newCaption:DWORD;FromDir:Boolean);
       procedure   TraceLinks;
      public
       constructor Create(Tree:PBPTree);
       destructor  Destroy; override;
       procedure   RenameLinks;
     end;

Implementation

function    TBPRename.GetTreeLevel(level:ShortString):LongInt; { Corrected }
Var ilev:Integer;
    i:Integer;
    s:String;
 begin
   Result:=-2;
   if length(level)=0 then Exit;
   Result:=StrToIntDef(level,-2); { Try -1,0  }
   if (Result<>-2) then
      begin
       if (Result=-1) or (Result=0) then
         Exit
        else
         begin
           Result:=1;
           Exit;
         end;
      end;
   { level is in x.x.x.x.x }
     ilev:=1;
     s:=level;
     i:=1;
     while (length(s)>0) do
      begin
        if (s[1]='.') then inc(ilev);
        s:=copy(s,2,length(s)-1);
      end;
     Result:=ilev;
 end;

constructor TBPRename.Create(Tree:PBPTree); { Corrected }
 begin
  FTree:=Tree;
  FNodesCount:=0;
  FLinksCount:=0;
  FNodes:=nil;
  FLinks:=nil;
  FMinLevel:=-1;
  FMaxLevel:=0;
  FCurLevel:=-1;
  FCurSSLevel:='-1';
  inherited Create;
 end;

destructor  TBPRename.Destroy;  { Corrected }
 begin
  FTree:=nil;
  inherited Destroy;
 end;

function  TBPRename.GetNextLevel(NodeNumber:DWORD):String; { Checked }
Var tmpNode:PBPNode;
    tmpTree:PBPTree;
    tmpNodesCount:integer;
    Flag:Boolean;
    i:integer;
    s:String;
 begin
   Result:='0';
   tmpTree:=FTree;
   Flag:=True;
   while (tmpTree<>nil) and (Flag) do
    begin
      if (tmpTree^.dwLevel=FCurLevel) then Flag:=False
      else tmpTree:=tmpTree^.next;
    end;
   if (not(Flag)) then
    begin
      tmpNode:=tmpTree^.lpNodes;
      tmpNodesCount:=tmpTree^.dwNodesCount;
      for i:=0 to (tmpNodesCount-1) do
       begin
         if (tmpNode^.dwNodeNumber=NodeNumber) then
          begin
           s:=tmpNode^.ssNodeNumber;
           s:=TrimLeft(s);
           s:=System.copy(s,2,Length(s)-1);
           Result:=s;
          end;
         inc(tmpNode);
       end;
    end;
 end;

procedure TBPRename.InitializeNodes(nodes:PBPNode;ncount:integer); { Corrected }
Var i:Integer;
    oldNode,tmpNode:PBPNodeRenameStructure;
    tmpNode1:PBPNode;
 begin
   FinalizeNodes;
   if (ncount=0) or (nodes=nil) then Exit;
   tmpNode1:=nodes;
   oldNode:=nil;
   for i:=0 to (ncount-1) do
    begin
      tmpNode:=PBPNodeRenameStructure(AllocMem(sizeof(TBPNodeRenameStructure)));
      tmpNode^.dwNodeNumber:=tmpNode1^.dwNodeNumber;
      tmpNode^.ssNodeNumber:=tmpNode1^.ssNodeNumber;
      tmpNode^.LinksLeft:=0;
      tmpNode^.LinksRight:=0;
      tmpNode^.LinksTop:=0;
      tmpNode^.LinksBottom:=0;
      tmpNode^.next:=nil;
      if (oldNode<>nil) then oldNode^.next:=tmpNode;
      oldNode:=tmpNode;
      if (i=0) then FNodes:=tmpNode;
      inc(tmpNode1);
    end;
    FNodesCount:=ncount;
 end;

procedure TBPRename.FinalizeNodes; { Corrected }
Var oldNode,tmpNode:PBPNodeRenameStructure;
 begin
   if (FNodes=nil) then Exit;
   tmpNode:=FNodes;
   while (tmpNode<>nil) do
    begin
      oldNode:=tmpNode;
      tmpNode:=tmpNode^.next;
      FreeMem(oldNode);
    end;
   FNodes:=nil;
   FNodesCount:=0;
 end;

function   TBPRename.GetMaxLevel:LongInt; { Corrected }
 Var tmpTree:PBPTree;
     lev:Integer;
  begin
    Result:=-2;
    if (FTree=nil) then Exit;
    tmpTree:=FTree;
    while (tmpTree<>nil) do
     begin
       lev:=GetTreeLevel(tmpTree^.ssLevel);
       if (lev=-1) or
          (lev=0) then
        begin
         if (lev>Result) then Result:=Lev;
        end
       else
        begin
         if (lev>Result) then
           Result:=lev;
        end;
       tmpTree:=tmpTree^.next;
     end;
  end;

 function    TBPRename.GetMinLevel(MaxLevel:LongInt):LongInt; { Corrected }
 Var tmpTree:PBPTree;
     lev:Integer;
  begin
    Result:=MaxLevel;
    if (FTree=nil) then Exit;
    tmpTree:=FTree;
    while (tmpTree<>nil) do
     begin
       lev:=GetTreeLevel(tmpTree^.ssLevel);
       if (Lev=-1) or
          (Lev=0) then
        begin
         if (Lev<Result) then Result:=Lev;
        end
       else
        begin
         if (Lev<Result) then
           Result:=Lev;
        end;
       tmpTree:=tmpTree^.next;
     end;
  end;

procedure TBPRename.InitializeLinks(links:PBPLink;lcount:integer); { Corrected }
Var i:Integer;
    oldLink,tmpLink:PBPLinkRenameStructure;
    tmpLink1:PBPLink;
 begin
   FinalizeLinks;
   if (lcount=0) or (links=nil) then Exit;
   tmpLink1:=links;
   oldLink:=nil;
   for i:=0 to (lcount-1) do
    begin
      tmpLink:=PBPLinkRenameStructure(AllocMem(sizeof(TBPLinkRenameStructure)));
      tmpLink^.link:=tmpLink1;
      tmpLink^.FromLinkNumber:=0;
      tmpLink^.ToLinkNumber:=0;
      tmpLink^.LinkRenamedFrom:=False;
      tmpLink^.LinkRenamedTo:=False;
      tmpLink^.next:=nil;
      if (oldLink<>nil) then oldLink^.next:=tmpLink;
      oldLink:=tmpLink;
      if (i=0) then FLinks:=tmpLink;
      inc(tmpLink1);
    end;
    FLinksCount:=lcount;
 end;

procedure TBPRename.FinalizeLinks; { Corrected }
Var oldLink,tmpLink:PBPLinkRenameStructure;
 begin
   if (FLinks=nil) then Exit;
   tmpLink:=FLinks;
   while (tmpLink<>nil) do
    begin
      oldLink:=tmpLink;
      tmpLink:=tmpLink^.next;
      oldLink^.link:=nil;
      FreeMem(oldLink);
    end;
   FLinks:=nil;
   FLinksCount:=0;
 end;

function  TBPRename.GetMultipleMask(index:Integer):TBPMultipleMask; { Checked }
 Var linkP:PBPLinkRenameStructure;
     linkC:PBPLinkRenameStructure;
     i:Integer;
     j:Integer;
  begin
   Result:=bpmm_None;
   if (index>=FLinksCount) or (index<0) then Exit;
   linkP:=FLinks;
   linkC:=FLinks;
   if (index>0) then
   begin
    j:=0;
    while (j<>index) do
       begin
        linkP:=linkP^.next;
        inc(j);
       end;
   end;
   for i:=0 to (FlinksCount-1) do
     begin
       if (i<>index) then
         begin
           if (linkC^.link^.dwFromNodeNumber=linkP^.link^.dwFromNodeNumber) and
              (linkC^.link^.dwFromLinkNumber=linkP^.link^.dwFromLinkNumber) and
              ((linkC^.link^.dwFlags and BP_FROMMASK)=(LinkP^.link^.dwFlags and BP_FROMMASK)) and
              (not(linkC^.linkRenamedFrom)) and
              ((linkC^.link^.dwToNodeNumber and BPGlobalNodeOperation)=0) and
              ((linkP^.link^.dwToNodeNumber and BPGlobalNodeOperation)=0) then
               begin
                 if Result=bpmm_From then Result:=bpmm_FromTo
                      else Result:=bpmm_To;
               end;
           if (linkC^.link^.dwToNodeNumber=linkP^.link^.dwToNodeNumber) and
              (linkC^.link^.dwToLinkNumber=linkP^.link^.dwToLinkNumber) and
              ((linkC^.link^.dwFlags and BP_TOMASK)=(LinkP^.link^.dwFlags and BP_TOMASK)) and
              (not(linkC^.linkRenamedTo))  and
              ((linkC^.link^.dwFromNodeNumber and BPGlobalNodeOperation)=0) and
              ((linkP^.link^.dwFromNodeNumber and BPGlobalNodeOperation)=0)  then
               begin
                 if Result=bpmm_To then Result:=bpmm_FromTo
                      else Result:=bpmm_From;
               end;
         end;
       linkC:=linkC^.next;
     end;
  end;


function TBPRename.IsMultipleFirst(index:Integer;MultipleFrom:Boolean):Boolean; { Checked }
Var tmpLink:PBPLinkRenameStructure;
    tmpLink1:PBPLinkRenameStructure;
    flag:Boolean;
    i,j:Integer;
 begin
   Result:=False;
   if (FLinksCount=0) then Exit;
   tmpLink:=FLinks;
   tmpLink1:=FLinks;
   if (index>0) then
   begin
    j:=0;
    while (j<>index) do
       begin
        tmpLink:=tmpLink^.next;
        inc(j);
       end;
   end;
   if ((tmpLink^.link^.dwFromNodeNumber and BPGlobalMask)=BPGlobalNodeOperation) and
      ((tmpLink^.link^.dwToNodeNumber and BPGlobalMask)=BPGlobalNodeOperation) then
       begin
         Result:=True;
         Exit;
       end;
   flag:=True;
   i:=0;
   while ((i<FLinksCount) and (flag)) do
    begin
      if MultipleFrom then
         begin
           if (tmpLink^.link^.dwToLinkNumber=tmpLink1^.link^.dwToLinkNumber) and
              (tmpLink^.link^.dwToNodeNumber=tmpLink1^.link^.dwToNodeNumber) and
              ((tmpLink^.link^.dwFlags and BP_TOMASK)=(tmpLink1^.link^.dwFlags and BP_TOMASK)) and
              ((tmpLink1^.link^.dwFromNodeNumber and BPGlobalMask)<>BPGlobalNodeOperation) and
              (not(tmpLink1^.linkRenamedTo))  then
             begin
               if (tmpLink=tmpLink1) then Result:=True;
               flag:=False;
             end;
         end
      else
         begin
           if (tmpLink^.link^.dwFromLinkNumber=tmpLink1^.link^.dwFromLinkNumber) and
              (tmpLink^.link^.dwFromNodeNumber=tmpLink1^.link^.dwFromNodeNumber) and
              ((tmpLink^.link^.dwFlags and BP_FROMMASK)=(tmpLink1^.link^.dwFlags and BP_FROMMASK)) and
              ((tmpLink1^.link^.dwToNodeNumber and BPGlobalMask)<>BPGlobalNodeOperation) and
              (not(tmpLink1^.linkRenamedFrom)) then
             begin
               if (tmpLink=tmpLink1) then Result:=True;
               flag:=False;
             end;
         end;
      tmpLink1:=tmpLink1^.next;
      inc(i);
    end;

 end;




procedure TBPRename.TraceLinks; { Checked }
Var i:integer;
    level:String;
    tmpLink:PBPLinkRenameStructure;
    tmpNode:PBPNodeRenameStructure;
    mm:TBPMultipleMask;
    mff,mft:Boolean;
    lt:TLinkType;
    ind:DWORD;
 begin
   for i:=0 to (FLinksCount-1) do
    begin { Main circle }
      tmpLink:=GetLink(i);    { Checked }
      mm:=GetMultipleMask(i);
      mff:=IsMultipleFirst(i,True);
      mft:=IsMultipleFirst(i,False);
      if (tmpLink<>nil) then
        begin { not null link }
          if (not(tmpLink^.linkRenamedFrom)) then
            begin   { link not renamed from }
              { From nodes }
              if ((tmpLink^.link^.dwFromNodeNumber and BPGlobalMask)=0) then
               begin { non-global from node }
                 if ((mm=bpmm_None) or (mm=bpmm_From)) then
                   begin { Simple link }
                     if ((tmpLink^.link^.dwToNodeNumber and BPGlobalNodeOperation)=0) then
                     begin
                     lt:=GetLinkTypeFrom(tmpLink^.link^.dwFlags);
                     case (lt) of
                                lt_left:  begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksLeft:=tmpNode^.linksLeft+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksLeft;
                                          end;
                                lt_right: begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksRight:=tmpNode^.linksRight+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksRight;
                                          end;
                                lt_top:   begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksTop:=tmpNode^.linksTop+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksTop;
                                          end;
                                lt_bottom:begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksBottom:=tmpNode^.linksBottom+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksBottom;
                                          end;
                     end; { end case }
                     if FCurSSLevel='-1' then level:='0' else
                     if FCurSSLevel='0' then level:=GetNextLevel(tmpLink^.link^.dwFromNodeNumber)
                     else level:=GetNextLevel(tmpLink^.link^.dwFromNodeNumber);
                     RenameGlobalLinks(level,tmpLink^.link^.dwFromLinkNumber
                       ,tmpLink^.FromLinkNumber,False,GetCheckLT(lt));
                     tmpLink^.link^.dwFromLinkNumber:=tmpLink^.FromLinkNumber;
                     tmpLink^.linkRenamedFrom:=True; { Ok, renamed }
                    end;
                   end   { Simple link }
                 else
                   begin { Multiple link }
                     if (mft) then { mm checked }
                       begin
                        if ((tmpLink^.link^.dwToNodeNumber and BPGlobalNodeOperation)=0) then
                        begin
                        lt:=GetLinkTypeFrom(tmpLink^.link^.dwFlags);
                        case (lt) of
                                lt_left:  begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksLeft:=tmpNode^.linksLeft+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksLeft;
                                          end;
                                lt_right: begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksRight:=tmpNode^.linksRight+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksRight;
                                          end;
                                lt_top:   begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksTop:=tmpNode^.linksTop+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksTop;
                                          end;
                                lt_bottom:begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwFromNodeNumber);
                                           tmpNode^.linksBottom:=tmpNode^.linksBottom+1;
                                           tmpLink^.FromLinkNumber:=tmpNode^.linksBottom;
                                          end;
                         end; { end case }
                        if FCurSSLevel='-1' then level:='0' else
                        if FCurSSLevel='0' then level:=GetNextLevel(tmpLink^.link^.dwFromNodeNumber)
                        else level:=GetNextLevel(tmpLink^.link^.dwFromNodeNumber);
                        RenameMultipleLinks(i,tmpLink^.FromLinkNumber,True);
                        RenameGlobalLinks(level,tmpLink^.link^.dwFromLinkNumber,
                                         tmpLink^.FromLinkNumber,False,GetCheckLT(lt));
                        tmpLink^.link^.dwFromLinkNumber:=tmpLink^.FromLinkNumber;
                        tmpLink^.linkRenamedFrom:=True; { Ok, renamed }
                       end { MultipleFirst }
                     end;
                   end;    { Multiple link }
               end;  { non-global from node }
            end; { link not renamed from }
          { To nodes }
          if ((tmpLink^.link^.dwToNodeNumber and BPGlobalMask)=0) then
          begin { non-global to node }
           if (not(tmpLink^.linkRenamedTo)) then
            begin   { link not renamed to }
             if ((mm=bpmm_None) or (mm=bpmm_To)) then
                   begin { Simple link }
                     if ((tmpLink^.link^.dwFromNodeNumber and BPGlobalNodeOperation)=0) then
                     begin
                     lt:=GetLinkTypeTo(tmpLink^.link^.dwFlags);
                     case (lt) of
                                lt_left:  begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksLeft:=tmpNode^.linksLeft+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksLeft;
                                          end;
                                lt_right: begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksRight:=tmpNode^.linksRight+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksRight;
                                          end;
                                lt_top:   begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksTop:=tmpNode^.linksTop+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksTop;
                                          end;
                                lt_bottom:begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksBottom:=tmpNode^.linksBottom+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksBottom;
                                          end;
                     end; { end case }
                     if FCurSSLevel='-1' then level:='0' else
                     if FCurSSLevel='0' then level:=GetNextLevel(tmpLink^.link^.dwToNodeNumber)
                     else level:=GetNextLevel(tmpLink^.link^.dwToNodeNumber);
                     RenameGlobalLinks(level,tmpLink^.link^.dwToLinkNumber,
                       tmpLink^.ToLinkNumber,True,GetCheckLT(lt));
                     tmpLink^.link^.dwToLinkNumber:=tmpLink^.ToLinkNumber;
                     tmpLink^.linkRenamedTo:=True; { Ok, renamed }
                    end;
                   end   { Simple link }
                 else
                   begin { Multiple link }
                     if (mff) then
                       begin
                        if ((tmpLink^.link^.dwFromNodeNumber and BPGlobalNodeOperation)=0) then
                        begin
                        lt:=GetLinkTypeTo(tmpLink^.link^.dwFlags);
                        case (lt) of
                                lt_left:  begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksLeft:=tmpNode^.linksLeft+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksLeft;
                                          end;
                                lt_right: begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksRight:=tmpNode^.linksRight+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksRight;
                                          end;
                                lt_top:   begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksTop:=tmpNode^.linksTop+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksTop;
                                          end;
                                lt_bottom:begin
                                           tmpNode:=GetNode(tmpLink^.link^.dwToNodeNumber);
                                           tmpNode^.linksBottom:=tmpNode^.linksBottom+1;
                                           tmpLink^.ToLinkNumber:=tmpNode^.linksBottom;
                                          end;
                         end; { end case }
                        if FCurSSLevel='-1' then level:='0' else
                        if FCurSSLevel='0' then level:=GetNextLevel(tmpLink^.link^.dwToNodeNumber)
                        else level:=GetNextLevel(tmpLink^.link^.dwToNodeNumber);
                        RenameMultipleLinks(i,tmpLink^.ToLinkNumber,False);
                        RenameGlobalLinks(level,tmpLink^.link^.dwToLinkNumber,
                          tmpLink^.ToLinkNumber,True,GetCheckLT(lt));
                        tmpLink^.link^.dwToLinkNumber:=tmpLink^.ToLinkNumber;
                        tmpLink^.linkRenamedTo:=True; { Ok, renamed }
                       end { MultipleFirst }
                     end;
                   end;    { Multiple link }
               end;  { non-global To node }
            end;    { link not renamed to }
        end;  { not null link }
    end;  { Main circle }
 end;

procedure TBPRename.RenameLinks;  { Checked }
Var i:LongInt;
    tmpTree:PBPTree;
    lcount,ncount:integer;
    links:PBPLink;
    nodes:PBPNode;
    lev:Integer;
 begin
   // WriteTree(FTree);
    FMaxLevel:=GetMaxLevel;
    if (FMaxLevel=-2) then Exit; { No levels found }
    FMinLevel:=GetMinLevel(FMaxLevel); { Corrected }
    for i:=FMinLevel to FMaxLevel do
     begin
      { Main circle for levels }
      tmpTree:=FTree;
      while (tmpTree<>nil) do
       begin
        lev:=GetTreeLevel(tmpTree^.ssLevel); { Corrected }
        if (lev=i) then
           begin
             { Find level for rename - rename links }
             lcount:=tmpTree^.dwLinksCount;
             links:=tmpTree^.lpLinks;
             nodes:=tmpTree^.lpNodes;
             ncount:=tmpTree^.dwNodesCount;
             if (lcount<>0) and (ncount<>0) then
              begin { Level }
                InitializeNodes(nodes,ncount); { Corrected }
                InitializeLinks(links,lcount); { Corrected }
                 { Let's rename }
                FCurSSLevel:=tmpTree^.ssLevel;
                FCurLevel:=tmpTree^.dwLevel;
                FParentLevel:=tmpTree^.dwParent;
                TraceLinks;
                FinalizeNodes;                 { Corrected }
                FinalizeLinks;                 { Corrected }
              end;  { Level }
           end; { Find level }
         tmpTree:=tmpTree^.next;
       end; { while }
     end; { Main circle for levels }
//    WriteTree(FTree);
 end;

function    TBPRename.GetNode(NodeNumber:DWORD):PBPNodeRenameStructure; { Checked }
Var tmpNode:PBPNodeRenameStructure;
    Flag:Boolean;
 begin
  Result:=nil;
  if (FNodesCount=0) then Exit;
  Flag:=True;
  tmpNode:=FNodes;
  while ((tmpNode<>nil) and (Flag)) do
    begin
      if tmpNode^.dwNodeNumber=NodeNumber then
        begin
          Result:=tmpNode;
          Flag:=False;
        end
      else tmpNode:=tmpNode^.next;
    end;
 end;

function    TBPRename.GetLink(index:integer):PBPLinkRenameStructure; { Corrected }
Var i:integer;
    tmpLink:PBPLinkRenameStructure;
 begin
  Result:=nil;
  if (FLinksCount=0) then Exit;
  tmpLink:=FLinks;
  for i:=0 to (index-1) do tmpLink:=tmpLink^.next;
  Result:=tmpLink;
 end;



function    TBPRename.GetLinkTypeFrom(Flags:DWORD):TLinkType;
  begin
    Result:=lt_Right;
    case (Flags and BP_FROMMASK) of
      BP_FROMNODE_LEFT:   Result:=lt_Left;
      BP_FROMNODE_RIGHT:  Result:=lt_Right;
      BP_FROMNODE_TOP:    Result:=lt_Top;
      BP_FROMNODE_BOTTOM: Result:=lt_Bottom;
    end;
  end;

function    TBPRename.GetLinkTypeTo(Flags:DWORD):TLinkType;
  begin
    Result:=lt_Left;
    case (Flags and BP_TOMASK) of
      BP_TONODE_LEFT:   Result:=lt_Left;
      BP_TONODE_RIGHT:  Result:=lt_Right;
      BP_TONODE_TOP:    Result:=lt_Top;
      BP_TONODE_BOTTOM: Result:=lt_Bottom;
    end;
  end;

function TBPRename.GetLocalNumber(Name:DWORD):TVeryShortString; { Checked }
 begin
   Result:=IntToStr(Name);
 end;

function TBPRename.GetCheckLT(lt:TLinkType):TLinkType; { Checked }
 begin
  case (lt) of
     lt_left:   Result:=lt_Right;
     lt_right:  Result:=lt_Left;
     lt_top:    Result:=lt_Bottom;
     lt_bottom: Result:=lt_Top;
  end;
 end;

procedure TBPRename.RenameGlobalLinks(level:String;oldName,newName:DWORD;FromDir:Boolean;check_lt:TLinkType); { checked }
Var tmpTree:PBPTree;
    Flag:Boolean;
    tmpLink:PBPLink;
    lcount:integer;
    i:integer;
    lt:TLinkType;
    newGlobal:TVeryShortString;
 begin
  if (FTree=nil) then Exit;
  tmpTree:=FTree;
  Flag:=True;
  while ((tmpTree<>nil) and (Flag)) do
   begin
    if (tmpTree^.ssLevel=level) then Flag:=False
    else tmpTree:=tmpTree^.next;
   end;
   if (tmpTree<>nil) then
     begin
       lcount:=tmpTree^.dwLinksCount;
       tmpLink:=tmpTree^.lpLinks;
       if (tmpLink<>nil) and (lcount<>0) then
         begin
           for i:=0 to (lcount-1) do
            begin
              case (FromDir) of
                True: begin
                        lt:=GetLinkTypeFrom(tmpLink^.dwFlags);
                        if (tmpLink^.dwFromLinkNumber=oldName) and
                           ((tmpLink^.dwFromNodeNumber and BPGlobalNotOpMask)<>0) and
                           (lt=check_lt)  then
                          begin
                            tmpLink^.dwFromLinkNumber:=newName;
                            case (lt) of
                             lt_left:   newGlobal:='O';
                             lt_right:  newGlobal:='I';
                             lt_top:    newGlobal:='M';
                             lt_bottom: newGlobal:='C';
                            end;
                           newGlobal:=newGlobal+GetLocalNumber(newName);
                           tmpLink^.ssGlobalCaption:=newGlobal;
                          end;
                      end;
                False: begin
                        lt:=GetLinkTypeTo(tmpLink^.dwFlags);
                        if (tmpLink^.dwToLinkNumber=oldName) and
                           ((tmpLink^.dwToNodeNumber and BPGlobalNotOpMask)<>0) and
                           (lt=check_lt) then
                          begin
                           tmpLink^.dwToLinkNumber:=newName;
                            case (lt) of
                             lt_left:   newGlobal:='O';
                             lt_right:  newGlobal:='I';
                             lt_top:    newGlobal:='M';
                             lt_bottom: newGlobal:='C';
                            end;
                           newGlobal:=newGlobal+GetLocalNumber(newName);
                           tmpLink^.ssGlobalCaption:=newGlobal;
                          end;
                      end;
              end;
              inc(tmpLink);
            end;
         end;
     end;
 end;
 

procedure TBPRename.RenameMultipleLinks(index:DWORD;newCaption:DWORD;FromDir:Boolean); { Checked }
Var tmpLink,tmpLink1:PBPLinkRenameStructure;
    i:DWORD;
 begin
  {}
  if (FLinksCount=0) then Exit;
  tmpLink1:=GetLink(index);
  for i:=0 to (FLinksCount-1) do
   begin
     tmpLink:=GetLink(i);
     if (i<>index) then
     begin
     case (FromDir) of
     True: begin
            if (tmpLink^.link^.dwFromNodeNumber=tmpLink1^.link^.dwFromNodeNumber) and
               (tmpLink^.link^.dwFromLinkNumber=tmpLink1^.link^.dwFromLinkNumber) and
               ((tmpLink^.link^.dwFlags and BP_FROMMASK)=(tmpLink1^.link^.dwFlags and BP_FROMMASK))
               then
               begin
                 tmpLink^.linkRenamedFrom:=True;
                 tmpLink^.link^.dwFromLinkNumber:=newCaption;
               end;
           end;
     False: begin
             if (tmpLink^.link^.dwToNodeNumber=tmpLink1^.link^.dwToNodeNumber) and
               (tmpLink^.link^.dwToLinkNumber=tmpLink1^.link^.dwToLinkNumber) and
               ((tmpLink^.link^.dwFlags and BP_TOMASK)=(tmpLink1^.link^.dwFlags and BP_TOMASK))
               then
               begin
                tmpLink^.linkRenamedTo:=True;
                tmpLink^.link^.dwToLinkNumber:=newCaption;
               end;
            end;
     end;
    end; { Not First Link }
   end; { For }
 end;

end.