{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 1996 AO ROSNO                   }
{         Copyright (c) 1997 Master-Bank                }
{                                                       }
{*******************************************************}

unit DotHold;

interface

uses SysUtils, Classes, DotDef,DotList;

{ TDotHolder }

type
  TDotHolder = class(TComponent)
  private
    FDot:    TDotList;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    procedure SetDot(Value: TDotList);
    procedure DotChanged(Sender: TObject);
    procedure DotChanging(Sender: TObject);
    procedure ReadDot(Reader: TReader);
    procedure WriteDot(Writer: TWriter);
  protected
    procedure AssignTo(Dest: TPersistent); override;
    procedure Changed; dynamic;
    procedure Changing; dynamic;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
  published
    property Dot:TDotList read FDot write SetDot stored False;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;

procedure Register;  

implementation

{ TDotHolder }

constructor TDotHolder.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDot := TDotList.Create;
  TDotList(FDot).OnChange := DotChanged;
  TDotList(FDot).OnChanging := DotChanging;
end;

destructor TDotHolder.Destroy;
begin
  FOnChange := nil;
  FOnChanging := nil;
  FDot.Free;
  inherited Destroy;
end;

procedure TDotHolder.AssignTo(Dest: TPersistent);
begin
  if Dest is TDotList then Dest.Assign(Dot)
  else inherited AssignTo(Dest);
end;

procedure TDotHolder.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDotHolder.Changing;
begin
  if Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TDotHolder.Clear;
begin
  FDot.Clear;
end;

procedure TDotHolder.DefineProperties(Filer: TFiler);

  function DoWrite: Boolean;
  var
    I: Integer;
    Ancestor: TDotHolder;
  begin
    Result:=False;
    Ancestor := TDotHolder(Filer.Ancestor);
    if (Ancestor <> nil) and (Ancestor.FDot.Count = FDot.Count) and
       (Ancestor.FDot.Count > 0) then
      for I := 1 to FDot.Count - 1 do begin
        Result := not(CompareMem(Pointer(Dot.Items[i]),
                               Pointer(Ancestor.FDot.Items[I]),
                               SizeOf(TDotStyle)));
        if Result then Break;
      end
    else Result := (FDot.Count > 0);
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('DotData', ReadDot, WriteDot, DoWrite);
end;

procedure TDotHolder.ReadDot(Reader: TReader);
begin
  Reader.ReadListBegin;
  FDot.Clear;
  Reader.ReadListEnd;
end;

procedure TDotHolder.SetDot(Value: TDotList);
begin
  FDot.Assign(Value);
end;

procedure TDotHolder.DotChanged(Sender: TObject);
begin
  if not (csReading in ComponentState) then Changed;
end;

procedure TDotHolder.DotChanging(Sender: TObject);
begin
  if not (csReading in ComponentState) then Changing;
end;

procedure TDotHolder.WriteDot(Writer: TWriter);
begin
  Writer.WriteListBegin;
  Writer.WriteListEnd;
end;

procedure Register;
 begin
   RegisterComponents('DeadSouls',[TDotHolder]);
 end;

end.