{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  :: QuickReport 2.0 for Delphi 1.0/2.0/3.0                  ::
  ::                                                         ::
  :: QRMERGE - Memo with expressions merger classes          ::
  :: This unit will be merged with TQRCtrls later            ::
  ::                                                         ::
  :: Copyright (c) 1997 QuSoft AS                            ::
  :: All Rights Reserved                                     ::
  ::                                                         ::
  :: web: http://www.qusoft.no    mail: support@qusoft.no    ::
  ::                              fax: +47 22 41 74 91       ::
  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

unit Qrmerge;

interface

uses sysutils, classes, qrprntr;

type
  TQRMerger = class
  private
    FOrgLines : TStrings;
    FStrippedLines : TStrings;
    FMergedLines : TStrings;
    FMerged : boolean;
    FPrepared : boolean;
    Expressions : TList;
    FDataSets : TList;
  protected
    function GetOrgLines : TStrings;
    function GetMergedLines : TStrings;
    procedure SetOrgLines(Value : TStrings);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Prepare;
    procedure Merge;
    procedure Cleanup;
    property Lines : TStrings read GetOrgLines write SetOrgLines;
    property MergedLines : TStrings read GetMergedLines;
    property Merged : boolean read FMerged;
    property Prepared : boolean read FPrepared;
    property DataSets :TList read FDataSets write FDataSets;
  end;

implementation

type
  TMemoEvaluator = class(TQREvaluator)
  public
    Line : integer;
    Position : integer;
  end;

constructor TQRMerger.Create;
begin
  FOrgLines := TStringList.Create;
  FMergedLines := nil;
  FMerged := false;
  FPrepared := false;
  Expressions := nil;
  FDataSets := TList.Create;
end;

destructor TQRMerger.Destroy;
begin
  if Prepared then Cleanup;
  if FOrgLines <> nil then
    FOrgLines.Free;
  if FDataSets <> nil then
    FDataSets.Free;
  inherited Destroy;
end;

function TQRMerger.GetOrgLines : TStrings;
begin
  Result := FOrgLines;
end;

function TQRMerger.GetMergedLines : TStrings;
begin
  if Merged then
    Result := FMergedLines
  else
    Result := nil;
end;

procedure TQRMerger.SetOrgLines(Value : TStrings);
begin
  if FOrgLines <> nil then
    FOrgLines.Free;
  FOrgLines := Value;
end;

procedure TQRMerger.Prepare;
var
  I, start, stop, Len : integer;
  Expr : String;
  aLine : string;
  aEvaluator : TMemoEvaluator;
begin
  if Prepared then Cleanup;
  if Lines.Count > 0 then
  begin
    Expressions := TList.Create;
    FStrippedLines := TStringList.Create;
    try
      for I := 0 to Lines.Count - 1 do
      begin
        aLine := FOrgLines[I];
        Start := Pos('{', aLine);
        while Start > 0 do
        begin
          stop := pos('}', aLine);
          Len := Stop - Start - 1;
          if Len > 0 then
          begin
            Expr := copy(aLine, start + 1, Len);
            Delete(aLine, Start, Len + 2);
            aEvaluator := TMemoEvaluator.Create;
            aEvaluator.DataSets := DataSets;
            aEvaluator.Prepare(Expr);
            aEvaluator.Line := I;
            aEvaluator.Position := Start;
            Expressions.Add(aEvaluator);
          end;
          Start := Pos('{', aLine);
        end;
        FStrippedLines.Add(aLine);
      end;
      FMergedLines := TStringList.Create;
    finally
      FPrepared := true;
    end;
  end;
end;

procedure TQRMerger.Cleanup;
var
  I : integer;
begin
  if Prepared then
  try
    for I := 0 to Expressions.Count - 1 do
      TMemoEvaluator(Expressions[I]).Free;
    Expressions.Free;
    FMergedLines.Free;
    FStrippedLines.Free;
  finally
    FPrepared := false;
    FMerged := false;
  end;
end;

procedure TQRMerger.Merge;
var
  I : integer;
  aResult : TQREvResult;
  Replacement : string;
  aLine : string;
begin
  if Expressions.Count > 0 then
  begin
    if Merged then
      FMergedLines.Clear;
    FMergedLines.Assign(FStrippedLines);
    for I := Expressions.Count - 1 downto 0 do
    begin
      with TMemoEvaluator(Expressions[I]) do
      begin
        aResult := Value;
        case aResult.Kind of
          resInt : Replacement := IntToStr(aResult.IntResult);
          resDouble : Replacement := FloatToStr(aResult.dblResult);
          resString : Replacement := aResult.StrResult;
          resError : Replacement := '';
        end;
        aLine := FMergedLines[Line];
        Insert(Replacement, aLine, Position);
        FMergedLines[Line] := aLine;
      end;
    end
  end else
    if not Merged then
      FMergedLines.Assign(FOrgLines);
  FMerged := true;
end;

end.
