{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 1995, 1996 AO ROSNO             }
{         Copyright (c) 1997 Master-Bank                }
{                                                       }
{*******************************************************}

unit RXCalc;

interface

uses {$IFDEF WIN32} Windows, {$ELSE} WinTypes, WinProcs, {$ENDIF} SysUtils,
  Messages, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Menus,
  ExtCtrls, Buttons, RxCtrls, Clipbrd;

const
  DefCalcPrecision = 15;

type
  TRxCalcState = (csFirst, csValid, csError);

{ TRxCalculatorForm }

  TRxCalculatorForm = class(TForm)
    MainPanel: TPanel;
    LabelPanel: TPanel;
    MemoryPanel: TPanel;
    DisplayPanel: TPanel;
    MemoryLabel: TLabel;
    DisplayLabel: TLabel;
    PopupMenu: TPopupMenu;
    CopyItem: TMenuItem;
    PasteItem: TMenuItem;
    OkButton: TRxSpeedButton;
    CancelButton: TRxSpeedButton;
    ButtonBack: TRxSpeedButton;
    ButtonC: TRxSpeedButton;
    ButtonMC: TRxSpeedButton;
    ButtonMR: TRxSpeedButton;
    ButtonMS: TRxSpeedButton;
    ButtonMP: TRxSpeedButton;
    Button0: TRxSpeedButton;
    Button1: TRxSpeedButton;
    Button2: TRxSpeedButton;
    Button3: TRxSpeedButton;
    Button4: TRxSpeedButton;
    Button5: TRxSpeedButton;
    Button6: TRxSpeedButton;
    Button7: TRxSpeedButton;
    Button8: TRxSpeedButton;
    Button9: TRxSpeedButton;
    ButtonPM: TRxSpeedButton;
    ButtonPnt: TRxSpeedButton;
    ButtonDiv: TRxSpeedButton;
    ButtonMul: TRxSpeedButton;
    ButtonSub: TRxSpeedButton;
    ButtonAdd: TRxSpeedButton;
    ButtonSqrt: TRxSpeedButton;
    ButtonPercent: TRxSpeedButton;
    ButtonRev: TRxSpeedButton;
    ButtonResult: TRxSpeedButton;
    procedure NumButtonClick(Sender: TObject);
    procedure ButtonPntClick(Sender: TObject);
    procedure ButtonDivClick(Sender: TObject);
    procedure ButtonMulClick(Sender: TObject);
    procedure ButtonSubClick(Sender: TObject);
    procedure ButtonAddClick(Sender: TObject);
    procedure ButtonResultClick(Sender: TObject);
    procedure ButtonCClick(Sender: TObject);
    procedure ButtonBackClick(Sender: TObject);
    procedure ButtonPMClick(Sender: TObject);
    procedure ButtonRevClick(Sender: TObject);
    procedure ButtonPercentClick(Sender: TObject);
    procedure ButtonSqrtClick(Sender: TObject);
    procedure ButtonMCClick(Sender: TObject);
    procedure ButtonMRClick(Sender: TObject);
    procedure ButtonMSClick(Sender: TObject);
    procedure ButtonMPClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure CopyItemClick(Sender: TObject);
    procedure PasteItemClick(Sender: TObject);
    procedure CheckFirst;
    procedure OkButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure PopupMenuPopup(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  private
    FStatus: TRxCalcState;
    FOperator: Char;
    FOperand: Double;
    FMemory: Double;
    FPrecision: Byte;
    FBeepOnError: Boolean;
    procedure CalcKey(Key: Char);
    procedure Clear;
    procedure Error;
    procedure SetDisplay(R: Double);
    function GetDisplay: Double;
    procedure UpdateMemoryLabel;
    function FindButton(Key: Char): TRxSpeedButton;
{$IFNDEF WIN32}
    procedure CreateParams(var Params: TCreateParams); override;
{$ENDIF}
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
  public
    property DisplayValue: Double read GetDisplay write SetDisplay;
  end;

{ TRxCalculator }

  TRxCalculator = class(TComponent)
  private
    FValue: Double;
    FMemory: Double;
    FCtl3D: Boolean;
    FPrecision: Byte;
    FBeepOnError: Boolean;
    FHelpContext: THelpContext;
    FCalc: TRxCalculatorForm;
    FOnChange: TNotifyEvent;
    FOnCalcKey: TKeyPressEvent;
    FOnDisplayChange: TNotifyEvent;
    function GetDisplay: Double;
  protected
    procedure Change; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    function Execute: Boolean;
    property CalcDisplay: Double read GetDisplay;
  published
    property BeepOnError: Boolean read FBeepOnError write FBeepOnError default True;
    property Ctl3D: Boolean read FCtl3D write FCtl3D default True;
    property HelpContext: THelpContext read FHelpContext write FHelpContext default 0;
    property Precision: Byte read FPrecision write FPrecision default DefCalcPrecision;
    property Value: Double read FValue write FValue;
    property OnCalcKey: TKeyPressEvent read FOnCalcKey write FOnCalcKey;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnDisplayChange: TNotifyEvent read FOnDisplayChange write FOnDisplayChange;
  end;

function CreateCalculatorForm(AOwner: TComponent; AHelpContext: THelpContext): TRxCalculatorForm;

implementation

uses {$IFNDEF WIN32} Str16, {$ENDIF} MaxMin, StrUtils;

{$R *.DFM}

function CreateCalculatorForm(AOwner: TComponent; AHelpContext: THelpContext): TRxCalculatorForm;
begin
  Result := TRxCalculatorForm.Create(AOwner);
  with Result do begin
    HelpContext := AHelpContext;
{$IFDEF WIN32}
    if HelpContext <> 0 then BorderIcons := BorderIcons + [biHelp];
{$ENDIF}
  end;
end;

{ TRxCalculator }

constructor TRxCalculator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCtl3D := True;
  FPrecision := DefCalcPrecision;
  FBeepOnError := True;
end;

function TRxCalculator.GetDisplay: Double;
begin
  if Assigned(FCalc) then Result := FCalc.GetDisplay
  else Result := FValue;
end;

procedure TRxCalculator.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

function TRxCalculator.Execute: Boolean;
begin
  FCalc := CreateCalculatorForm(Self, HelpContext);
  with FCalc do
  try
    Ctl3D := FCtl3D;
    FMemory := Self.FMemory;
    UpdateMemoryLabel;
    FPrecision := Max(2, Self.Precision);
    FBeepOnError := Self.BeepOnError;
    if Self.FValue <> 0 then begin
      DisplayValue := Self.FValue;
      FStatus := csFirst;
      FOperator := '=';
    end;
    Result := (ShowModal = mrOk);
    if Result then begin
      Self.FMemory := FMemory;
      if (DisplayValue <> Self.FValue) then begin
        Self.FValue := DisplayValue;
        Change;
      end;
    end;
  finally
    Free;
    FCalc := nil;
  end;
end;

{ TRxCalculatorForm }

procedure TRxCalculatorForm.Error;
begin
  FStatus := csError;
  DisplayLabel.Caption := 'Error';
  if FBeepOnError then MessageBeep(0);
end;

{$IFNDEF WIN32}
procedure TRxCalculatorForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if Application.MainForm <> nil then
    Params.WndParent := Application.MainForm.Handle;
end;
{$ENDIF WIN32}

procedure TRxCalculatorForm.SetDisplay(R: Double);
var
  S: string;
begin
  S := FloatToStrF(R, ffGeneral, Max(2, FPrecision), 0);
  if DisplayLabel.Caption <> S then begin
    DisplayLabel.Caption := S;
    if (Owner <> nil) and (Owner is TRxCalculator) then
      if Assigned(TRxCalculator(Owner).FOnDisplayChange) then
        TRxCalculator(Owner).FOnDisplayChange(Owner);
  end;
end;

function TRxCalculatorForm.GetDisplay: Double;
begin
  if FStatus = csError then Result := 0.0
  else Result := StrToFloat(Trim(DisplayLabel.Caption));
end;

procedure TRxCalculatorForm.CheckFirst;
begin
  if FStatus = csFirst then begin
    FStatus := csValid;
    DisplayLabel.Caption := '0';
  end;
end;

procedure TRxCalculatorForm.CMCtl3DChanged(var Message: TMessage);
const
  Ctl3DStyle: array[Boolean] of TButtonStyle = (bsWin31, bsNew);
  Ctl3DBevel: array[Boolean] of TPanelBevel = (bvNone, bvLowered);
var
  I: Integer;
begin
  inherited;
  for I := 0 to ComponentCount - 1 do begin
    if Components[I] is TRxSpeedButton then
      TRxSpeedButton(Components[I]).Style := Ctl3DStyle[Ctl3D]
    else if Components[I] = MemoryPanel then
      MemoryPanel.BevelInner := Ctl3DBevel[Ctl3D]
    else if Components[I] = DisplayPanel then
      DisplayPanel.BevelOuter := Ctl3DBevel[Ctl3D]
    else if Components[I] = MainPanel then
      MainPanel.BevelOuter := Ctl3DBevel[Ctl3D];
  end;
end;

procedure TRxCalculatorForm.UpdateMemoryLabel;
begin
  if FMemory <> 0.0 then MemoryLabel.Caption := 'M'
  else MemoryLabel.Caption := '';
end;

procedure TRxCalculatorForm.CalcKey(Key: Char);
var
  R: Double;
begin
  Key := UpCase(Key);
  if (FStatus = csError) and (Key <> 'C') then Key := #0;
  if (Owner <> nil) and (Owner is TRxCalculator) then
    if Assigned(TRxCalculator(Owner).FOnCalcKey) then
      TRxCalculator(Owner).FOnCalcKey(Owner, Key);
  if Key in [DecimalSeparator, '.', ','] then begin
    CheckFirst;
    if Pos(DecimalSeparator, DisplayLabel.Caption) = 0 then
      DisplayLabel.Caption := DisplayLabel.Caption + DecimalSeparator;
    Exit;
  end;
  case Key of
    'R':
      if FStatus in [csValid, csFirst] then begin
        FStatus := csFirst;
        if GetDisplay = 0 then Error else SetDisplay(1.0 / GetDisplay);
      end;
    'Q':
      if FStatus in [csValid, csFirst] then begin
        FStatus := csFirst;
        if GetDisplay < 0 then Error else SetDisplay(Sqrt(GetDisplay));
      end;
    '0'..'9':
      begin
        CheckFirst;
        if DisplayLabel.Caption = '0' then DisplayLabel.Caption := '';
        if Pos('E', DisplayLabel.Caption) = 0 then begin
          if Length(DisplayLabel.Caption) < Max(2, FPrecision) +
            Ord(Boolean(Pos('-', DisplayLabel.Caption))) then
            DisplayLabel.Caption := DisplayLabel.Caption + Key
          else if FBeepOnError then MessageBeep(0);
        end;
      end;
    #8:
      begin
        CheckFirst;
        if (Length(DisplayLabel.Caption) = 1) or
          ((Length(DisplayLabel.Caption) = 2) and
          (DisplayLabel.Caption[1] = '-'))
        then
          DisplayLabel.Caption := '0'
        else
          DisplayLabel.Caption := Copy(DisplayLabel.Caption, 1,
            Length(DisplayLabel.Caption) - 1);
      end;
    '_': SetDisplay(-GetDisplay);
    '+', '-', '*', '/', '=', '%', #13:
      begin
        if FStatus = csValid then begin
          FStatus := csFirst;
          R := GetDisplay;
          if Key = '%' then
            case FOperator of
              '+', '-': R := FOperand * R / 100.0;
              '*', '/': R := R / 100.0;
            end;
          case FOperator of
            '+': SetDisplay(FOperand + R);
            '-': SetDisplay(FOperand - R);
            '*': SetDisplay(FOperand * R);
            '/': if R = 0 then Error else SetDisplay(FOperand / R);
          end;
        end;
        FOperator := Key;
        FOperand := GetDisplay;
      end;
    #27, 'C': Clear;
    ^C: CopyItemClick(nil);
    ^V: PasteItemClick(nil);
  end;
end;

procedure TRxCalculatorForm.Clear;
begin
  FStatus := csFirst;
  SetDisplay(0.0);
  FOperator := '=';
end;

procedure TrxCalculatorForm.FormKeyPress(Sender: TObject; var Key: Char);
var
  Btn: TRxSpeedButton;
begin
  Btn := FindButton(Key);
  if Btn <> nil then begin
    Btn.ButtonClick;
  end else CalcKey(Key);
end;

function TRxCalculatorForm.FindButton(Key: Char): TRxSpeedButton;
const
  ButtonChars = '0123456789_./*-+Q%R='#8'C';
var
  I: Integer;
  BtnTag: Longint;
begin
  if Key in [DecimalSeparator, '.', ','] then Key := '.'
  else if Key = #13 then Key := '='
  else if Key = #27 then Key := 'C';
  BtnTag := Pos(UpCase(Key), ButtonChars) - 1;
  if BtnTag >= 0 then
    for I := 0 to MainPanel.ControlCount - 1 do begin
      if (MainPanel.Controls[I] is TRxSpeedButton) then begin
        Result := TRxSpeedButton(MainPanel.Controls[I]);
        if Result.Tag = BtnTag then Exit;
      end;
    end;
  Result := nil;
end;

procedure TRxCalculatorForm.NumButtonClick(Sender: TObject);
begin
  CalcKey(Char(TComponent(Sender).Tag + Ord('0')));
end;

procedure TRxCalculatorForm.ButtonPntClick(Sender: TObject);
begin
  CalcKey(DecimalSeparator);
end;

procedure TRxCalculatorForm.ButtonDivClick(Sender: TObject);
begin
  CalcKey('/');
end;

procedure TRxCalculatorForm.ButtonMulClick(Sender: TObject);
begin
  CalcKey('*');
end;

procedure TRxCalculatorForm.ButtonSubClick(Sender: TObject);
begin
  CalcKey('-');
end;

procedure TRxCalculatorForm.ButtonAddClick(Sender: TObject);
begin
  CalcKey('+');
end;

procedure TRxCalculatorForm.ButtonResultClick(Sender: TObject);
begin
  CalcKey('=');
end;

procedure TRxCalculatorForm.ButtonCClick(Sender: TObject);
begin
  CalcKey('C');
end;

procedure TRxCalculatorForm.ButtonBackClick(Sender: TObject);
begin
  CalcKey(#8);
end;

procedure TRxCalculatorForm.ButtonPMClick(Sender: TObject);
begin
  CalcKey('_');
end;

procedure TRxCalculatorForm.ButtonPercentClick(Sender: TObject);
begin
  CalcKey('%');
end;

procedure TRxCalculatorForm.ButtonRevClick(Sender: TObject);
begin
  CalcKey('R');
end;

procedure TRxCalculatorForm.ButtonSqrtClick(Sender: TObject);
begin
  CalcKey('Q');
end;

procedure TRxCalculatorForm.ButtonMCClick(Sender: TObject);
begin
  FMemory := 0.0;
  UpdateMemoryLabel;
end;

procedure TRxCalculatorForm.ButtonMRClick(Sender: TObject);
begin
  if FStatus in [csValid, csFirst] then begin
    FStatus := csFirst;
    CheckFirst;
    SetDisplay(FMemory);
  end;
end;

procedure TRxCalculatorForm.ButtonMSClick(Sender: TObject);
begin
  if FStatus in [csValid, csFirst] then begin
    FStatus := csFirst;
    FMemory := GetDisplay;
    UpdateMemoryLabel;
  end;
end;

procedure TRxCalculatorForm.ButtonMPClick(Sender: TObject);
begin
  if FStatus in [csValid, csFirst] then begin
    FStatus := csFirst;
    FMemory := FMemory + GetDisplay;
    UpdateMemoryLabel;
  end;
end;

procedure TRxCalculatorForm.FormCreate(Sender: TObject);
begin
  FMemory := 0.0;
  FPrecision := DefCalcPrecision;
  FBeepOnError := True;
end;

procedure TRxCalculatorForm.CopyItemClick(Sender: TObject);
begin
  Clipboard.AsText := DisplayLabel.Caption;
end;

procedure TRxCalculatorForm.PasteItemClick(Sender: TObject);
begin
  if Clipboard.HasFormat(CF_TEXT) then
    try
      SetDisplay(StrToFloat(Trim(ReplaceStr(Clipboard.AsText,
        CurrencyString, ''))));
    except
      DisplayLabel.Caption := '0';
    end;
end;

procedure TRxCalculatorForm.OkButtonClick(Sender: TObject);
begin
  if FStatus <> csError then begin
    DisplayValue := DisplayValue; { to raise exception on error }
    ModalResult := mrOk;
  end
  else if FBeepOnError then MessageBeep(0);
end;

procedure TRxCalculatorForm.CancelButtonClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TRxCalculatorForm.PopupMenuPopup(Sender: TObject);
begin
  PasteItem.Enabled := Clipboard.HasFormat(CF_TEXT);
end;

end.
