Unit Debug;
{
(c) .  1997,2006
}
{$INCLUDE Compmode.inc}
Interface

Uses Windows,SysUtils,BPDef,classes;
procedure OpenLog;
procedure CloseLog;
procedure WriteLog(const logstring:ShortString);
procedure WriteNode(node:PBPNode);
procedure WriteNodes(StartNode:TList);
procedure WriteLink(link:PBPLink);
procedure WriteLinks(StartLink:TList);
procedure WriteLevel(tree:PBPTree);
procedure WriteTree(tree,IPNodes,OPTree,LTypes,NTypes:TList;onsave:TLevelEvent);
procedure WriteLinkTypes(LTypes:TList);
procedure WriteNodeTypes(NTypes:TList);
procedure WriteExternal(IPNodes:TList);
procedure WriteDB(OpTree:TList);

Const filename='bpd.txt';

Implementation
var loglist:TStringList;
procedure OpenLog;
begin
loglist:=TstringList.Create;
loglist.Add('/!-'+datetimetostr(now));
end;
procedure closelog;
var logold:TstringList;
begin
loglist.Add('/!!'+datetimetostr(now));
logold:=TStringList.Create;
if fileexists(filename) then
logold.loadfromfile(filename);
logold.addstrings(loglist);
logold.SaveToFile(filename);
logold.free;
loglist.Free;
end;
procedure WriteLog(const logstring:ShortString);
 begin
loglist.add(logstring);
 end;

procedure WriteNode(node:PBPNode);
Var s:ShortString;
 begin
    WriteLog('');
    WriteLog('NodeBegin');
    if (node<>nil) then
     begin
       s:=node^.ssCaption;
       WriteLog('Caption: '+s);
       s:=IntToStr(node^.dwNodeNumber);
       WriteLog('Number: '+s);
       s:=node^.ssNodeNumber;
       WriteLog('DrawNumber: '+s);
     end
    else WriteLog('EmptyNode');
    WriteLog('NodeEnd');
    WriteLog('');
 end;

procedure WriteNodes(StartNode:TList);
Var i:integer;
 begin
   for i:=0 to StartNode.count-1 do
      WriteNode(StartNode[i]);
 end;

procedure WriteLink(link:PBPLink);
 begin
   WriteLog('');
   WriteLog('LinkBegin');
    if (link<>nil) then
     begin
        if ((link^.dwFlags and BP_SHOW_FROMCAPTION)<>0) then
        WriteLog('FromCaption: '+link^.ssFromCaption);
        if ((link^.dwFlags and BP_SHOW_TOCAPTION)<>0) then
        WriteLog('ToCaption: '+link^.ssToCaption);
        if ((link^.dwFlags and BP_SHOW_GLOBALCAPTION)<>0) then
        WriteLog('GlobalCaption: '+link^.ssGlobalCaption);
        if ((link^.dwFromNodeNumber and BPGlobalMask)=0) then
         WriteLog('FromNode: '+IntToStr(link^.dwFromNodeNumber))
        else
         begin
          case (link^.dwFromNodeNumber and BPGlobalMask) of
           BPGlobalNodeLeft:  WriteLog('FromGlobalNumber: GlobalLeft');
           BPGlobalNodeRight: WriteLog('FromGlobalNumber: GlobalRight');
           BPGlobalNodeTop:   WriteLog('FromGlobalNumber: GlobalTop');
           BPGlobalNodeBottom:WriteLog('FromGlobalNumber: GlobalBottom');
           BPGlobalNodeOperation: WriteLog('FromGlobalNumber: Operation');
           BPGlobalNodeInProcess: WriteLog('FromGlobalNumber: ExternalProcess');
          end;
          WriteLog('FromParentNode: '+IntToStr(link^.dwFromNodeNumber and BPNotGlobalMask));
         end;
         if link^.IP_SRC>0 then
         Writelog('FromExternal: '+inttostr(link^.IP_SRC));
        case (link^.dwFlags and BP_FROMMASK) of
          BP_FROMNODE_LEFT:    WriteLog('FromSide: Left');
          BP_FROMNODE_RIGHT:   WriteLog('FromSide: Right');
          BP_FROMNODE_TOP:     WriteLog('FromSide: Top');
          BP_FROMNODE_BOTTOM:  WriteLog('FromSide: Bottom');
        end;
        WriteLog('FromICAM: '+IntToStr(link^.dwFromLinkNumber));
        if ((link^.dwToNodeNumber and BPGlobalMask)=0) then
         WriteLog('ToNode: '+IntToStr(link^.dwToNodeNumber))
        else
         begin
          case (link^.dwToNodeNumber and BPGlobalMask) of
           BPGlobalNodeLeft:  WriteLog('ToGlobalNumber: GlobalLeft');
           BPGlobalNodeRight: WriteLog('ToGlobalNumber: GlobalRight');
           BPGlobalNodeTop:   WriteLog('ToGlobalNumber: GlobalTop');
           BPGlobalNodeBottom:WriteLog('ToGlobalNumber: GlobalBottom');
           BPGlobalNodeOperation: WriteLog('ToGlobalNumber: Operation');
           BPGlobalNodeInProcess: WriteLog('ToGlobalNumber: ExternalProcess');
          end;
          WriteLog('ToParentNode: '+IntToStr(link^.dwToNodeNumber and BPNotGlobalMask));
         end;
         if link^.IP_DST>0 then
         Writelog('ToExternal: '+inttostr(link^.IP_DST));

        case (link^.dwFlags and BP_TOMASK) of
          BP_TONODE_LEFT:    WriteLog('ToSide: Left');
          BP_TONODE_RIGHT:   WriteLog('ToSide: Right');
          BP_TONODE_TOP:     WriteLog('ToSide: Top');
          BP_TONODE_BOTTOM:  WriteLog('ToSide: Bottom');
        end;
        WriteLog('ToICAM: '+IntToStr(link^.dwToLinkNumber));
        if ((link^.dwFlags and BP_NOTUSED_FROM)<>0) then
        WriteLog('NotUsed: From');
        if ((link^.dwFlags and BP_NOTUSED_TO)<>0) then
        WriteLog('NotUsed: To');
     end
    else WriteLog('EmptyLink');
    WriteLog('LinkEnd');
    WriteLog('');
 end;

procedure WriteLinks(StartLink:TList);
Var i:integer;
 begin
    for i:=0 to StartLink.count-1  do
       WriteLink(StartLink[i]);
 end;

procedure WriteLevel(tree:PBPTree);
 begin
   WriteLog('');
   WriteLog('LevelBegin');
    if (tree<>nil) then
     begin
       WriteLog('Caption: '+tree^.ssCaption);
       WriteLog('Level: '+IntToStr(tree^.dwLevel));
       WriteLog('SSLevel: '+tree^.ssLevel);
       WriteLog('ParentLevel: '+IntToStr(tree^.dwParent));
       WriteNodes(tree^.lpNodes);
       WriteLinks(tree^.lpLinks);
     end
    else
     WriteLog('EmptyLevel');
   WriteLog('LevelEnd');
   WriteLog('');
 end;

procedure WriteTree(tree,IPNodes,OPTree,LTypes,NTypes:TList;onsave:TLevelEvent);
Var   i:integer;
 begin
   OpenLog;
   WriteLinkTypes(LTypes);
   WriteNodeTypes(NTypes);
   WriteExternal(IPNodes);
   WriteDB(OpTree);
   WriteLog('TreeBegin');
   for i:=0   to tree.count-1 do
        begin
          if assigned(onsave) then onsave(PBPTree(Tree[i])^.ssCaption);
          WriteLevel(PBPTree(Tree[i]));
        end;
   WriteLog('TreeEnd');
   CloseLog;
 end;
procedure WriteLinkTypes(LTypes:TList);
var i:integer;
    lt:PBPLinkTypes;
begin
WriteLog('<Link Types>');
 for i:=0 to LTypes.Count-1 do
 begin
 lt:=PBPLinkTypes(LTypes[i]);
 writeLog(' TypeBegin');
 writeLog('  Caption='+lt^.ssCaption);
 writeLog('  NodeIndex='+inttostr(lt^.dwLinkIndex));
 writeLog('  ID='+inttostr(lt^.dwID));
 writeLog('  Parent='+inttostr(lt^.dwParent));
 writeLog(' TypeEnd');
 end;
WriteLog('</Link Types>');
end;
procedure WriteNodeTypes(NTypes:TList);
var i:integer;
    nt:PBPNodeTypes;
begin
WriteLog('<Node Types>');
 for i:=0 to NTypes.Count-1 do
 begin
 nt:=PBPNodeTypes(NTypes[i]);
 writeLog(' TypeBegin');
 writeLog('  Caption='+nt^.ssCaption);
 writeLog('  NodeIndex='+inttostr(nt^.dwNodeIndex));
 writeLog('  ID='+inttostr(nt^.dwID));
 writeLog('  Parent='+inttostr(nt^.dwParent));
 writeLog(' TypeEnd');
 end;
WriteLog('</Node Types>');
end;
procedure WriteExternal(IPNodes:TList);
var i:integer;
    ip:PBPInProcessNode;
begin
WriteLog('<External Nodes>');
for i:=0 to IPNodes.count-1 do
begin
 ip:=PBPInProcessNode(IPNodes[i]);
 writeLog(' IPBegin');
 WriteLog('  Caption='+ip^.ssCaption);
 WriteLog('  Index='+inttostr(ip^.dwNodeIndex));
 WriteLog('  Number='+inttostr(ip^.dwNodeNumber));
 writeLog(' IpEnd');
end;
WriteLog('</External Nodes>');
end;

procedure WriteDB(OpTree:TList);
var i:integer;
    db:PBPOPTree;
begin
WriteLog('<DB Nodes>');
 for i:=0 to OpTree.Count-1 do
 begin
  db:=PBPOPTree(OpTree[i]);
  WriteLog(' DB_Begin');
  writeLog('  Caption='+db^.ssCaption);
  writeLog('  Level='+IntTostr(db^.dwLevel));
  WriteLog(' DB_End');
 end;
WriteLog('</DB Nodes>');
end;

end.
